import asyncio
import json
from collections.abc import Awaitable, Callable
from dataclasses import dataclass
from typing import Any

from blackgeorge.adapters.base import ModelResponse
from blackgeorge.core.event import Event
from blackgeorge.core.job import Job
from blackgeorge.core.message import Message
from blackgeorge.core.pending_action import PendingAction
from blackgeorge.core.report import Report
from blackgeorge.core.tool_call import ToolCall
from blackgeorge.core.types import RunStatus
from blackgeorge.store.state import RunState
from blackgeorge.tools.base import Tool, ToolResult
from blackgeorge.tools.execution import aexecute_tool
from blackgeorge.utils import new_id
from blackgeorge.worker_context import (
    SUMMARY_ATTEMPT_LIMIT,
    context_error_message,
)
from blackgeorge.worker_messages import (
    emit_assistant_message,
    replace_tool_call,
    structured_content,
    tool_call_with_result,
    tool_message,
)
from blackgeorge.worker_tools import (
    pending_options,
    tool_action_type,
    tool_prompt,
)

EventEmitter = Callable[[str, str, dict[str, Any]], None]


@dataclass(frozen=True)
class ToolPlan:
    ordered_calls: list[ToolCall]
    executable_calls: list[tuple[ToolCall, Tool]]
    immediate_results: dict[str, ToolResult]
    pending: PendingAction | None
    max_tool_calls_exceeded: bool


@dataclass(frozen=True)
class ContextDecision:
    retry: bool
    report: Report | None


def _build_report(
    run_id: str,
    status: RunStatus,
    content: str | None = None,
    reasoning_content: str | None = None,
    data: Any | None = None,
    messages: list[Message] | None = None,
    tool_calls: list[ToolCall] | None = None,
    metrics: dict[str, Any] | None = None,
    events: list[Event] | None = None,
    pending_action: PendingAction | None = None,
    errors: list[str] | None = None,
) -> Report:
    return Report(
        run_id=run_id,
        status=status,
        content=content,
        reasoning_content=reasoning_content,
        data=data,
        messages=list(messages) if messages else [],
        tool_calls=list(tool_calls) if tool_calls else [],
        metrics=metrics or {},
        events=list(events) if events else [],
        pending_action=pending_action,
        errors=list(errors) if errors else [],
    )


def _build_state(
    run_id: str,
    status: RunStatus,
    runner_name: str,
    job: Job,
    messages: list[Message],
    tool_calls: list[ToolCall],
    pending_action: PendingAction | None,
    metrics: dict[str, Any],
    iteration: int,
    payload: dict[str, Any] | None = None,
) -> RunState:
    return RunState(
        run_id=run_id,
        status=status,
        runner_type="worker",
        runner_name=runner_name,
        job=job,
        messages=messages,
        tool_calls=tool_calls,
        pending_action=pending_action,
        metrics=metrics,
        iteration=iteration,
        payload=payload or {},
    )


def _report_error(
    run_id: str,
    messages: list[Message],
    errors: list[str],
    events: list[Event],
) -> Report:
    return Report(
        run_id=run_id,
        status="failed",
        content=None,
        reasoning_content=None,
        data=None,
        messages=messages,
        tool_calls=[],
        metrics={},
        events=events,
        pending_action=None,
        errors=errors,
    )


def _ensure_not_running_loop(action: str, async_action: str) -> None:
    try:
        asyncio.get_running_loop()
    except RuntimeError:
        return
    raise RuntimeError(
        f"{action} cannot be called from a running event loop. Use {async_action} instead."
    )


def _fail_report(
    *,
    run_id: str,
    worker_name: str,
    message: str,
    messages: list[Message],
    tool_calls: list[ToolCall],
    metrics: dict[str, Any],
    events: list[Event],
    errors: list[str],
    emit: EventEmitter,
) -> Report:
    errors.append(message)
    emit("worker.failed", worker_name, {"error": message})
    return _build_report(
        run_id,
        "failed",
        None,
        None,
        None,
        messages,
        tool_calls,
        metrics,
        events,
        None,
        errors,
    )


async def _acontext_retry(
    *,
    run_id: str,
    worker_name: str,
    messages: list[Message],
    tool_calls: list[ToolCall],
    metrics: dict[str, Any],
    events: list[Event],
    errors: list[str],
    emit: EventEmitter,
    model_registered: bool,
    respect_context_window: bool,
    context_summaries: int,
    apply_summary: Callable[[], Awaitable[bool]],
) -> ContextDecision:
    if not respect_context_window:
        message = context_error_message(model_registered, False)
        return ContextDecision(
            False,
            _fail_report(
                run_id=run_id,
                worker_name=worker_name,
                message=message,
                messages=messages,
                tool_calls=tool_calls,
                metrics=metrics,
                events=events,
                errors=errors,
                emit=emit,
            ),
        )
    if context_summaries >= SUMMARY_ATTEMPT_LIMIT:
        message = context_error_message(model_registered, True)
        return ContextDecision(
            False,
            _fail_report(
                run_id=run_id,
                worker_name=worker_name,
                message=message,
                messages=messages,
                tool_calls=tool_calls,
                metrics=metrics,
                events=events,
                errors=errors,
                emit=emit,
            ),
        )
    if not await apply_summary():
        message = context_error_message(model_registered, True)
        return ContextDecision(
            False,
            _fail_report(
                run_id=run_id,
                worker_name=worker_name,
                message=message,
                messages=messages,
                tool_calls=tool_calls,
                metrics=metrics,
                events=events,
                errors=errors,
                emit=emit,
            ),
        )
    return ContextDecision(True, None)


def _should_stream(stream: bool, _tools: list[Tool], response_schema: Any | None) -> bool:
    return stream and response_schema is None


def _tool_result_preview(result: ToolResult, limit: int) -> tuple[str | None, bool]:
    if result.content is not None:
        text = result.content
    elif result.data is not None:
        try:
            text = json.dumps(result.data, ensure_ascii=True)
        except (TypeError, ValueError):
            text = str(result.data)
    elif result.error is not None:
        text = result.error
    else:
        return None, False
    if len(text) > limit:
        return f"{text[:limit]}...", True
    return text, False


def _tool_event_payload(call: ToolCall, result: ToolResult, limit: int = 200) -> dict[str, Any]:
    payload: dict[str, Any] = {"tool_call_id": call.id}
    preview, truncated = _tool_result_preview(result, limit)
    if preview is not None:
        payload["result_preview"] = preview
        payload["result_truncated"] = truncated
    if result.timed_out:
        payload["timed_out"] = True
    if result.cancelled:
        payload["cancelled"] = True
    return payload


def _record_usage(metrics: dict[str, Any], response: object) -> None:
    if isinstance(response, ModelResponse):
        metrics["usage"] = response.usage
    else:
        metrics["usage"] = {}


def _finalize_structured_response(
    *,
    run_id: str,
    data: Any,
    messages: list[Message],
    tool_calls: list[ToolCall],
    metrics: dict[str, Any],
    events: list[Event],
    errors: list[str],
    emit: EventEmitter,
    worker_name: str,
) -> Report:
    content = structured_content(data)
    assistant_message = Message(role="assistant", content=content)
    messages.append(assistant_message)
    emit_assistant_message(emit, worker_name, assistant_message)
    emit("worker.completed", worker_name, {})
    return _build_report(
        run_id,
        "completed",
        content,
        None,
        data,
        messages,
        tool_calls,
        metrics,
        events,
        None,
        errors,
    )


def _finalize_plain_response(
    *,
    run_id: str,
    response: ModelResponse,
    messages: list[Message],
    tool_calls: list[ToolCall],
    metrics: dict[str, Any],
    events: list[Event],
    errors: list[str],
    emit: EventEmitter,
    worker_name: str,
) -> Report:
    assistant_message = Message(
        role="assistant",
        content=response.content or "",
        reasoning_content=response.reasoning_content,
        thinking_blocks=response.thinking_blocks,
    )
    messages.append(assistant_message)
    emit_assistant_message(emit, worker_name, assistant_message)
    emit("worker.completed", worker_name, {})
    return _build_report(
        run_id,
        "completed",
        response.content,
        response.reasoning_content,
        None,
        messages,
        tool_calls,
        metrics,
        events,
        None,
        errors,
    )


def _plan_tool_calls(
    *,
    response: ModelResponse,
    allowed_tools: dict[str, Tool],
    tool_calls: list[ToolCall],
    max_tool_calls: int,
) -> ToolPlan:
    ordered_calls: list[ToolCall] = []
    executable_calls: list[tuple[ToolCall, Tool]] = []
    immediate_results: dict[str, ToolResult] = {}
    pending: PendingAction | None = None
    max_tool_calls_exceeded = False

    for call in response.tool_calls:
        if len(tool_calls) >= max_tool_calls:
            max_tool_calls_exceeded = True
            break
        tool_calls.append(call)
        if call.error:
            ordered_calls.append(call)
            immediate_results[call.id] = ToolResult(error=call.error)
            continue

        tool = allowed_tools.get(call.name)
        if tool is None:
            ordered_calls.append(call)
            immediate_results[call.id] = ToolResult(error=f"Tool not found: {call.name}")
            continue
        action_type = tool_action_type(tool)
        if action_type:
            metadata = {"tool": tool.name}
            if tool.input_key:
                metadata["input_key"] = tool.input_key
            pending = PendingAction(
                action_id=new_id(),
                type=action_type,
                tool_call=call,
                prompt=tool_prompt(tool, action_type, call),
                options=pending_options(action_type),
                metadata=metadata,
            )
            break
        ordered_calls.append(call)
        executable_calls.append((call, tool))

    return ToolPlan(
        ordered_calls=ordered_calls,
        executable_calls=executable_calls,
        immediate_results=immediate_results,
        pending=pending,
        max_tool_calls_exceeded=max_tool_calls_exceeded,
    )


async def _execute_tool_calls_async(
    ordered_calls: list[ToolCall],
    executable_calls: list[tuple[ToolCall, Tool]],
    immediate_results: dict[str, ToolResult],
    messages: list[Message],
    tool_calls: list[ToolCall],
    emit: EventEmitter,
) -> None:
    results: dict[str, ToolResult] = dict(immediate_results)
    if executable_calls:
        for call, tool in executable_calls:
            emit("tool.started", tool.name, {"tool_call_id": call.id})
        if len(executable_calls) == 1:
            call, tool = executable_calls[0]
            results[call.id] = await aexecute_tool(tool, call)
        else:
            tasks = [aexecute_tool(tool, call) for call, tool in executable_calls]
            tool_results = await asyncio.gather(*tasks)
            for (call, _), result in zip(executable_calls, tool_results, strict=True):
                results[call.id] = result
    for call in ordered_calls:
        result = results.get(call.id, ToolResult(error="Tool execution failed"))
        if result.error:
            emit("tool.failed", call.name, {"tool_call_id": call.id, "error": result.error})
        else:
            emit("tool.completed", call.name, _tool_event_payload(call, result))
        tool_result_message = tool_message(result, call)
        messages.append(tool_result_message)
        replace_tool_call(tool_calls, tool_call_with_result(call, result))



from collections.abc import Callable
from typing import Any

from blackgeorge.adapters.base import BaseModelAdapter
from blackgeorge.core.event import Event
from blackgeorge.core.job import Job
from blackgeorge.core.report import Report
from blackgeorge.store.state import RunState
from blackgeorge.tools.base import Tool
from blackgeorge.tools.registry import Toolbelt
from blackgeorge.worker_runner import WorkerRunner

EventEmitter = Callable[[str, str, dict[str, Any]], None]


class Worker:
    def __init__(
        self,
        name: str,
        tools: list[Tool] | None = None,
        model: str | None = None,
        instructions: str | None = None,
        memory_scope: str | None = None,
    ) -> None:
        self.name = name
        self.model = model
        self.instructions = instructions
        self.toolbelt = Toolbelt(tools)
        self.memory_scope = memory_scope or f"worker:{name}"

    def tools(self) -> list[Tool]:
        return self.toolbelt.list()

    def _runner(self) -> WorkerRunner:
        return WorkerRunner(self.name, self.toolbelt, self.instructions)

    def run(
        self,
        *,
        adapter: BaseModelAdapter,
        job: Job,
        run_id: str,
        events: list[Event],
        emit: EventEmitter,
        temperature: float | None,
        max_tokens: int | None,
        stream: bool,
        stream_options: dict[str, Any] | None,
        structured_output_retries: int,
        max_iterations: int,
        max_tool_calls: int,
        model_name: str,
        respect_context_window: bool = True,
    ) -> tuple[Report, RunState | None]:
        return self._runner().run(
            adapter=adapter,
            job=job,
            run_id=run_id,
            events=events,
            emit=emit,
            temperature=temperature,
            max_tokens=max_tokens,
            stream=stream,
            stream_options=stream_options,
            structured_output_retries=structured_output_retries,
            max_iterations=max_iterations,
            max_tool_calls=max_tool_calls,
            model_name=model_name,
            respect_context_window=respect_context_window,
        )

    async def arun(
        self,
        *,
        adapter: BaseModelAdapter,
        job: Job,
        run_id: str,
        events: list[Event],
        emit: EventEmitter,
        temperature: float | None,
        max_tokens: int | None,
        stream: bool,
        stream_options: dict[str, Any] | None,
        structured_output_retries: int,
        max_iterations: int,
        max_tool_calls: int,
        model_name: str,
        respect_context_window: bool = True,
    ) -> tuple[Report, RunState | None]:
        return await self._runner().arun(
            adapter=adapter,
            job=job,
            run_id=run_id,
            events=events,
            emit=emit,
            temperature=temperature,
            max_tokens=max_tokens,
            stream=stream,
            stream_options=stream_options,
            structured_output_retries=structured_output_retries,
            max_iterations=max_iterations,
            max_tool_calls=max_tool_calls,
            model_name=model_name,
            respect_context_window=respect_context_window,
        )

    def resume(
        self,
        *,
        adapter: BaseModelAdapter,
        state: RunState,
        decision_or_input: Any,
        events: list[Event],
        emit: EventEmitter,
        temperature: float | None,
        max_tokens: int | None,
        stream: bool,
        stream_options: dict[str, Any] | None,
        structured_output_retries: int,
        max_iterations: int,
        max_tool_calls: int,
        model_name: str,
        respect_context_window: bool = True,
    ) -> tuple[Report, RunState | None]:
        return self._runner().resume(
            adapter=adapter,
            state=state,
            decision_or_input=decision_or_input,
            events=events,
            emit=emit,
            temperature=temperature,
            max_tokens=max_tokens,
            stream=stream,
            stream_options=stream_options,
            structured_output_retries=structured_output_retries,
            max_iterations=max_iterations,
            max_tool_calls=max_tool_calls,
            model_name=model_name,
            respect_context_window=respect_context_window,
        )

    async def aresume(
        self,
        *,
        adapter: BaseModelAdapter,
        state: RunState,
        decision_or_input: Any,
        events: list[Event],
        emit: EventEmitter,
        temperature: float | None,
        max_tokens: int | None,
        stream: bool,
        stream_options: dict[str, Any] | None,
        structured_output_retries: int,
        max_iterations: int,
        max_tool_calls: int,
        model_name: str,
        respect_context_window: bool = True,
    ) -> tuple[Report, RunState | None]:
        return await self._runner().aresume(
            adapter=adapter,
            state=state,
            decision_or_input=decision_or_input,
            events=events,
            emit=emit,
            temperature=temperature,
            max_tokens=max_tokens,
            stream=stream,
            stream_options=stream_options,
            structured_output_retries=structured_output_retries,
            max_iterations=max_iterations,
            max_tool_calls=max_tool_calls,
            model_name=model_name,
            respect_context_window=respect_context_window,
        )



import hashlib
import json
import logging
import re
from typing import Any, cast

import chromadb
import numpy as np
from chromadb.api.types import Documents, Embeddable, EmbeddingFunction, Embeddings, Metadatas
from chromadb.config import DEFAULT_DATABASE, DEFAULT_TENANT
from chromadb.errors import NotFoundError
from numpy.typing import NDArray

from blackgeorge.memory.base import MemoryScope, MemoryStore
from blackgeorge.utils import utc_now

logger = logging.getLogger(__name__)

DEFAULT_CHUNK_SIZE = 8000
DEFAULT_CHUNK_OVERLAP = 200
DEFAULT_EMBEDDING_DIMENSIONS = 256
TOKEN_PATTERN = re.compile(r"[a-z0-9]+")


def _chunk_text(
    text: str,
    chunk_size: int = DEFAULT_CHUNK_SIZE,
    overlap: int = DEFAULT_CHUNK_OVERLAP,
) -> list[str]:
    if len(text) <= chunk_size:
        return [text]
    if chunk_size <= 0:
        return [text]
    chunks: list[str] = []
    start = 0
    step = max(chunk_size - overlap, 1)
    while start < len(text):
        end = min(start + chunk_size, len(text))
        chunks.append(text[start:end])
        if end == len(text):
            break
        start += step
    return chunks


def _serialize_value(value: Any) -> str:
    if isinstance(value, str):
        return value
    return json.dumps(value, ensure_ascii=True, default=str)


def _deserialize_value(text: str) -> Any:
    try:
        return json.loads(text)
    except json.JSONDecodeError:
        return text


class DeterministicEmbeddingFunction(EmbeddingFunction[Embeddable]):
    def __init__(self, dimensions: int = DEFAULT_EMBEDDING_DIMENSIONS) -> None:
        self._dimensions = max(int(dimensions), 1)

    def __call__(self, input: Embeddable) -> Embeddings:
        if not input:
            return []
        first_item = input[0]
        if isinstance(first_item, str):
            documents = cast(Documents, input)
            return [self._embed_document(document) for document in documents]
        raise TypeError("DeterministicEmbeddingFunction only supports text documents")

    @staticmethod
    def name() -> str:
        return "blackgeorge_deterministic_embedding"

    @staticmethod
    def build_from_config(config: dict[str, Any]) -> "DeterministicEmbeddingFunction":
        dimensions_raw = config.get("dimensions", DEFAULT_EMBEDDING_DIMENSIONS)
        try:
            dimensions = int(dimensions_raw)
        except (TypeError, ValueError):
            dimensions = DEFAULT_EMBEDDING_DIMENSIONS
        return DeterministicEmbeddingFunction(dimensions=dimensions)

    def get_config(self) -> dict[str, Any]:
        return {"dimensions": self._dimensions}

    def _embed_document(self, document: str) -> NDArray[np.float32]:
        vector = np.zeros(self._dimensions, dtype=np.float32)
        tokens = TOKEN_PATTERN.findall(document.lower())
        if not tokens:
            tokens = [document.lower()] if document else [""]
        for token in tokens:
            digest = hashlib.blake2b(token.encode("utf-8"), digest_size=8).digest()
            hashed = int.from_bytes(digest, byteorder="big", signed=False)
            index = hashed % self._dimensions
            direction = 1.0 if ((hashed >> 1) & 1) == 0 else -1.0
            vector[index] += direction
        magnitude = float(np.linalg.norm(vector))
        if magnitude == 0.0:
            vector[0] = 1.0
            return vector
        return vector / magnitude


def _deterministic_embedding_from_collection(
    collection: Any,
) -> DeterministicEmbeddingFunction | None:
    model = getattr(collection, "_model", None)
    if model is None:
        return None
    configuration_json = getattr(model, "configuration_json", None)
    if not isinstance(configuration_json, dict):
        return None
    embedding_function_config = configuration_json.get("embedding_function")
    if not isinstance(embedding_function_config, dict):
        return None
    name = embedding_function_config.get("name")
    if name != DeterministicEmbeddingFunction.name():
        return None
    config = embedding_function_config.get("config", {})
    if not isinstance(config, dict):
        config = {}
    dimensions_raw = config.get("dimensions", DEFAULT_EMBEDDING_DIMENSIONS)
    try:
        dimensions = int(dimensions_raw)
    except (TypeError, ValueError):
        dimensions = DEFAULT_EMBEDDING_DIMENSIONS
    return DeterministicEmbeddingFunction(dimensions=dimensions)


class VectorMemoryStore(MemoryStore):
    def __init__(
        self,
        path: str,
        collection_name: str = "memories",
        tenant: str = DEFAULT_TENANT,
        database: str = DEFAULT_DATABASE,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        chunk_overlap: int = DEFAULT_CHUNK_OVERLAP,
        embedding_function: EmbeddingFunction[Embeddable] | None = None,
    ) -> None:
        self._path = path
        self._embedding_function = embedding_function
        self._client = chromadb.PersistentClient(
            path=path,
            tenant=tenant,
            database=database,
        )
        self._collection_name = collection_name
        self._chunk_size = max(chunk_size, 1)
        self._chunk_overlap = max(chunk_overlap, 0)
        try:
            existing_collection = self._client.get_collection(name=collection_name)
            if self._embedding_function is None:
                self._embedding_function = _deterministic_embedding_from_collection(
                    existing_collection
                )
            if self._embedding_function is None:
                self._collection = existing_collection
            else:
                self._collection = self._client.get_collection(
                    name=collection_name,
                    embedding_function=self._embedding_function,
                )
        except NotFoundError:
            if self._embedding_function is None:
                self._embedding_function = DeterministicEmbeddingFunction()
            self._collection = self._client.get_or_create_collection(
                name=collection_name,
                metadata={"hnsw:space": "cosine"},
                embedding_function=self._embedding_function,
            )

    def _resolve_chunk_config(self, metadata: dict[str, Any]) -> tuple[int, int]:
        chunk_size_raw = metadata.get("chunk_size", self._chunk_size)
        chunk_overlap_raw = metadata.get("chunk_overlap", self._chunk_overlap)
        try:
            chunk_size = int(chunk_size_raw)
        except (TypeError, ValueError):
            chunk_size = self._chunk_size
        try:
            chunk_overlap = int(chunk_overlap_raw)
        except (TypeError, ValueError):
            chunk_overlap = self._chunk_overlap
        if chunk_size <= 0:
            chunk_size = self._chunk_size
        if chunk_overlap < 0:
            chunk_overlap = 0
        return chunk_size, chunk_overlap

    def write(self, key: str, value: Any, scope: MemoryScope) -> None:
        text = _serialize_value(value)
        chunks = _chunk_text(text, chunk_size=self._chunk_size, overlap=self._chunk_overlap)
        now = utc_now().isoformat()
        self._delete_by_key(key, scope)
        ids: list[str] = []
        metadatas: Metadatas = []
        for i, _ in enumerate(chunks):
            ids.append(f"{scope}:{key}:{i}")
            metadatas.append(
                {
                    "scope": scope,
                    "key": key,
                    "chunk_index": i,
                    "total_chunks": len(chunks),
                    "chunk_size": self._chunk_size,
                    "chunk_overlap": self._chunk_overlap,
                    "created_at": now,
                }
            )
        self._collection.add(
            ids=ids,
            documents=chunks,
            metadatas=metadatas,
        )

    def read(self, key: str, scope: MemoryScope) -> Any | None:
        results = self._collection.get(
            where={"$and": [{"scope": {"$eq": scope}}, {"key": {"$eq": key}}]},  # type: ignore[dict-item]
            include=["documents", "metadatas"],
        )
        if not results["ids"]:
            return None
        docs = results["documents"] or []
        metadatas = results["metadatas"] or []
        pairs: list[tuple[dict[str, Any], str]] = []
        for meta, doc in zip(metadatas, docs, strict=False):
            pairs.append((dict(meta), str(doc)))
        sorted_chunks = sorted(pairs, key=lambda x: int(x[0].get("chunk_index", 0)))
        chunk_meta = sorted_chunks[0][0] if sorted_chunks else {}
        chunk_size, chunk_overlap = self._resolve_chunk_config(chunk_meta)
        step = max(chunk_size - chunk_overlap, 1)
        overlap_size = chunk_size - step
        text_parts: list[str] = []
        for idx, (_, doc) in enumerate(sorted_chunks):
            if idx == 0:
                text_parts.append(doc)
            else:
                new_len = len(doc) - overlap_size
                if new_len > 0:
                    text_parts.append(doc[-new_len:])
        full_text = "".join(text_parts)
        return _deserialize_value(full_text)

    def search(
        self,
        query: str,
        scope: MemoryScope,
        top_k: int = 5,
    ) -> list[tuple[str, Any]]:
        results = self._collection.query(
            query_texts=[query],
            where={"scope": {"$eq": scope}},  # type: ignore[dict-item]
            n_results=top_k,
            include=["documents", "metadatas", "distances"],
        )
        if not results["ids"] or not results["ids"][0]:
            return []
        seen_keys: set[str] = set()
        matches: list[tuple[str, Any]] = []
        ids = results["ids"][0]
        metadatas = results["metadatas"][0] if results["metadatas"] else []
        for i, doc_id in enumerate(ids):
            meta = dict(metadatas[i]) if i < len(metadatas) else {}
            key = str(meta.get("key", doc_id))
            if key in seen_keys:
                continue
            seen_keys.add(key)
            value = self.read(key, scope)
            if value is not None:
                matches.append((key, value))
        return matches

    def reset(self, scope: MemoryScope) -> None:
        try:
            self._collection.delete(
                where={"scope": {"$eq": scope}},  # type: ignore[dict-item]
            )
        except Exception as exc:
            logger.debug("Failed to reset memory scope %s: %s", scope, exc)

    def _delete_by_key(self, key: str, scope: MemoryScope) -> None:
        try:
            self._collection.delete(
                where={"$and": [{"scope": {"$eq": scope}}, {"key": {"$eq": key}}]},  # type: ignore[dict-item]
            )
        except Exception as exc:
            logger.debug("Failed to delete memory key %s in scope %s: %s", key, scope, exc)



from typing import Any

from blackgeorge.core.pending_action import PendingAction
from blackgeorge.core.tool_call import ToolCall
from blackgeorge.core.types import PendingActionType
from blackgeorge.tools.base import Tool


def tool_prompt(tool: Tool, action_type: str, tool_call: ToolCall) -> str:
    if action_type == "user_input":
        question = tool_call.arguments.get("question")
        if isinstance(question, str) and question.strip():
            return question
        return tool.user_input_prompt or tool.description or f"Provide input for {tool.name}"
    base = tool.confirmation_prompt or tool.description or f"Confirm {tool.name}"
    path = tool_call.arguments.get("path")
    if isinstance(path, str) and path.strip():
        return f"{base} ({path})"
    return base


def tool_action_type(tool: Tool) -> PendingActionType | None:
    if tool.requires_confirmation:
        return "confirmation"
    if tool.requires_user_input:
        return "user_input"
    return None


def pending_options(action_type: PendingActionType) -> list[str]:
    if action_type == "confirmation":
        return ["yes", "no"]
    return []


def resume_argument_key(pending_action: PendingAction) -> str:
    value = pending_action.metadata.get("input_key")
    return value if isinstance(value, str) else "user_input"


def update_arguments(call: ToolCall, key: str, value: Any) -> ToolCall:
    updated = dict(call.arguments)
    updated[key] = value
    return call.model_copy(update={"arguments": updated})



import json
import os
from typing import TYPE_CHECKING, Any, Literal

from pydantic import BaseModel

from blackgeorge.adapters.base import BaseModelAdapter
from blackgeorge.adapters.litellm import LiteLLMAdapter
from blackgeorge.core.event import Event
from blackgeorge.core.job import Job
from blackgeorge.core.message import Message
from blackgeorge.core.report import Report
from blackgeorge.event_bus import EventBus
from blackgeorge.memory.in_memory import InMemoryMemoryStore
from blackgeorge.store.base import RunStore
from blackgeorge.store.sqlite import SQLiteRunStore
from blackgeorge.utils import new_id, utc_now
from blackgeorge.worker import Worker
from blackgeorge.workflow.flow import Flow
from blackgeorge.workforce import Workforce

if TYPE_CHECKING:
    from blackgeorge.session import WorkerSession


class Desk:
    def __init__(
        self,
        *,
        model: str,
        temperature: float | None = None,
        max_tokens: int | None = None,
        stream: bool = False,
        structured_stream_mode: Literal["off", "preview"] = "off",
        structured_output_retries: int = 3,
        max_iterations: int = 10,
        max_tool_calls: int = 20,
        respect_context_window: bool = True,
        event_bus: EventBus | None = None,
        run_store: RunStore | None = None,
        memory_store: Any | None = None,
        adapter: BaseModelAdapter | None = None,
        storage_dir: str | None = None,
    ) -> None:
        self.model = model
        self.temperature = temperature
        self.max_tokens = max_tokens
        self.stream = stream
        if structured_stream_mode not in ("off", "preview"):
            raise ValueError("structured_stream_mode must be 'off' or 'preview'")
        self.structured_stream_mode = structured_stream_mode
        self.structured_output_retries = structured_output_retries
        self.max_iterations = max_iterations
        self.max_tool_calls = max_tool_calls
        self.respect_context_window = respect_context_window
        self.event_bus = event_bus or EventBus()
        self.adapter = adapter or LiteLLMAdapter()
        self.storage_dir = storage_dir or ".blackgeorge"
        os.makedirs(self.storage_dir, exist_ok=True)
        self.db_path = os.path.join(self.storage_dir, "blackgeorge.db")
        self.run_store = run_store or SQLiteRunStore(self.db_path)
        self.memory_store = memory_store or InMemoryMemoryStore()
        self._workers: dict[str, Worker] = {}
        self._workforces: dict[str, Workforce] = {}
        self._flow_runs: dict[str, Flow] = {}

    def register_worker(self, worker: Worker) -> None:
        self._workers[worker.name] = worker

    def register_workforce(self, workforce: Workforce) -> None:
        self._workforces[workforce.name] = workforce

    def unregister_worker(self, worker: Worker | str) -> None:
        name = worker if isinstance(worker, str) else worker.name
        self._workers.pop(name, None)

    def unregister_workforce(self, workforce: Workforce | str) -> None:
        name = workforce if isinstance(workforce, str) else workforce.name
        self._workforces.pop(name, None)

    def register_flow_run(self, run_id: str, flow: Flow) -> None:
        self._flow_runs[run_id] = flow

    def unregister_flow_run(self, run_id: str) -> None:
        self._flow_runs.pop(run_id, None)

    def flow(self, steps: list[Any], name: str | None = None) -> Flow:
        return Flow(self, steps, name=name)

    def session(
        self,
        worker: Worker,
        *,
        session_id: str | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> "WorkerSession | None":
        from blackgeorge.session import WorkerSession

        if session_id:
            return WorkerSession.resume(session_id=session_id, worker=worker, desk=self)
        return WorkerSession.start(
            worker=worker, desk=self, session_id=session_id, metadata=metadata
        )

    def _emit(
        self,
        events: list[Event],
        run_id: str,
        event_type: str,
        source: str,
        payload: dict[str, Any],
    ) -> None:
        event = Event(
            event_id=new_id(),
            type=event_type,
            timestamp=utc_now(),
            run_id=run_id,
            source=source,
            payload=payload,
        )
        events.append(event)
        self.event_bus.emit(event)
        self.run_store.add_event(event)

    def _output_json(self, report: Report) -> Any | None:
        if isinstance(report.data, BaseModel):
            return report.data.model_dump(mode="json", warnings=False)
        return report.data

    def _apply_memory(self, worker: Worker, job: Job) -> Job:
        if self.memory_store is None:
            return job
        memory_value = self.memory_store.read("context", worker.memory_scope)
        if memory_value is None:
            return job
        if isinstance(memory_value, str):
            content = memory_value
        else:
            content = json.dumps(memory_value, ensure_ascii=True, default=str)
        memory_message = Message(role="system", content=f"Memory:\n{content}")
        if job.initial_messages:
            messages = list(job.initial_messages)
            insert_index = 0
            while insert_index < len(messages) and messages[insert_index].role == "system":
                insert_index += 1
            messages.insert(insert_index, memory_message)
        else:
            messages = [memory_message]
        return job.model_copy(update={"initial_messages": messages})

    def _resolve_structured_stream_mode(self, job: Job) -> Job:
        if job.structured_stream_mode is not None:
            return job
        return job.model_copy(update={"structured_stream_mode": self.structured_stream_mode})

    def _write_memory(self, worker: Worker, report: Report) -> None:
        if self.memory_store is None:
            return
        if report.status != "completed":
            return
        value: Any | None = report.data if report.data is not None else report.content
        if value is None:
            return
        if isinstance(value, BaseModel):
            value = value.model_dump(mode="json", warnings=False)
        self.memory_store.write("last_output", value, worker.memory_scope)

    def run(
        self,
        runner: Worker | Workforce,
        job: Job,
        *,
        stream: bool | None = None,
        run_id: str | None = None,
    ) -> Report:
        run_id = run_id or new_id()
        events: list[Event] = []
        stream_enabled = self.stream if stream is None else stream
        stream_options = {"include_usage": True} if stream_enabled else None
        job = self._resolve_structured_stream_mode(job)
        if isinstance(runner, Worker):
            job = self._apply_memory(runner, job)
        self.run_store.create_run(run_id, job.model_dump(mode="json"))
        self._emit(events, run_id, "run.started", "desk", {"job_id": job.id})

        def emit(event_type: str, source: str, payload: dict[str, Any]) -> None:
            self._emit(events, run_id, event_type, source, payload)

        if isinstance(runner, Worker):
            self.register_worker(runner)
            report, state = runner.run(
                adapter=self.adapter,
                job=job,
                run_id=run_id,
                events=events,
                emit=emit,
                temperature=self.temperature,
                max_tokens=self.max_tokens,
                stream=stream_enabled,
                stream_options=stream_options,
                structured_output_retries=self.structured_output_retries,
                max_iterations=self.max_iterations,
                max_tool_calls=self.max_tool_calls,
                model_name=runner.model or self.model,
                respect_context_window=self.respect_context_window,
            )
        elif isinstance(runner, Workforce):
            self.register_workforce(runner)
            report, state = runner.run(
                adapter=self.adapter,
                job=job,
                run_id=run_id,
                events=events,
                emit=emit,
                temperature=self.temperature,
                max_tokens=self.max_tokens,
                stream=stream_enabled,
                stream_options=stream_options,
                structured_output_retries=self.structured_output_retries,
                max_iterations=self.max_iterations,
                max_tool_calls=self.max_tool_calls,
                default_model=self.model,
                respect_context_window=self.respect_context_window,
                drain_async_handlers=self.event_bus.await_pending,
            )
        else:
            raise TypeError("Runner must be Worker or Workforce")

        if state is not None and report.status == "paused":
            state.payload["stream"] = stream_enabled
            self._emit(events, run_id, "run.paused", "desk", {})
            self.run_store.update_run(run_id, "paused", report.content, None, state)
        elif report.status == "completed":
            self._emit(events, run_id, "run.completed", "desk", {})
            self.run_store.update_run(
                run_id,
                "completed",
                report.content,
                self._output_json(report),
                None,
            )
            if isinstance(runner, Worker):
                self._write_memory(runner, report)
        else:
            self._emit(events, run_id, "run.failed", "desk", {"errors": report.errors})
            self.run_store.update_run(
                run_id,
                "failed",
                report.content,
                self._output_json(report),
                None,
            )

        return report

    async def arun(
        self,
        runner: Worker | Workforce,
        job: Job,
        *,
        stream: bool | None = None,
        run_id: str | None = None,
    ) -> Report:
        run_id = run_id or new_id()
        events: list[Event] = []
        stream_enabled = self.stream if stream is None else stream
        stream_options = {"include_usage": True} if stream_enabled else None
        job = self._resolve_structured_stream_mode(job)
        if isinstance(runner, Worker):
            job = self._apply_memory(runner, job)
        self.run_store.create_run(run_id, job.model_dump(mode="json"))
        self._emit(events, run_id, "run.started", "desk", {"job_id": job.id})

        def emit(event_type: str, source: str, payload: dict[str, Any]) -> None:
            self._emit(events, run_id, event_type, source, payload)

        if isinstance(runner, Worker):
            self.register_worker(runner)
            report, state = await runner.arun(
                adapter=self.adapter,
                job=job,
                run_id=run_id,
                events=events,
                emit=emit,
                temperature=self.temperature,
                max_tokens=self.max_tokens,
                stream=stream_enabled,
                stream_options=stream_options,
                structured_output_retries=self.structured_output_retries,
                max_iterations=self.max_iterations,
                max_tool_calls=self.max_tool_calls,
                model_name=runner.model or self.model,
                respect_context_window=self.respect_context_window,
            )
        elif isinstance(runner, Workforce):
            self.register_workforce(runner)
            report, state = await runner.arun(
                adapter=self.adapter,
                job=job,
                run_id=run_id,
                events=events,
                emit=emit,
                temperature=self.temperature,
                max_tokens=self.max_tokens,
                stream=stream_enabled,
                stream_options=stream_options,
                structured_output_retries=self.structured_output_retries,
                max_iterations=self.max_iterations,
                max_tool_calls=self.max_tool_calls,
                default_model=self.model,
                respect_context_window=self.respect_context_window,
            )
        else:
            raise TypeError("Runner must be Worker or Workforce")

        if state is not None and report.status == "paused":
            state.payload["stream"] = stream_enabled
            self._emit(events, run_id, "run.paused", "desk", {})
            self.run_store.update_run(run_id, "paused", report.content, None, state)
        elif report.status == "completed":
            self._emit(events, run_id, "run.completed", "desk", {})
            self.run_store.update_run(
                run_id,
                "completed",
                report.content,
                self._output_json(report),
                None,
            )
            if isinstance(runner, Worker):
                self._write_memory(runner, report)
        else:
            self._emit(events, run_id, "run.failed", "desk", {"errors": report.errors})
            self.run_store.update_run(
                run_id,
                "failed",
                report.content,
                self._output_json(report),
                None,
            )

        return report

    def resume(
        self,
        report: Report,
        decision_or_input: Any,
        *,
        stream: bool | None = None,
    ) -> Report:
        record = self.run_store.get_run(report.run_id)
        if record is None:
            failed = Report(
                run_id=report.run_id,
                status="failed",
                content=None,
                reasoning_content=None,
                data=None,
                messages=report.messages,
                tool_calls=report.tool_calls,
                metrics=report.metrics,
                events=report.events,
                pending_action=None,
                errors=["No stored state"],
            )
            return failed
        events = self.run_store.get_events(report.run_id)

        def resume_failed(message: str) -> Report:
            failed = Report(
                run_id=report.run_id,
                status="failed",
                content=None,
                data=None,
                messages=report.messages,
                tool_calls=report.tool_calls,
                metrics=report.metrics,
                events=events,
                pending_action=None,
                errors=[message],
            )
            self._emit(
                events,
                report.run_id,
                "run.failed",
                "desk",
                {"errors": failed.errors},
            )
            self.run_store.update_run(
                report.run_id,
                "failed",
                failed.content,
                self._output_json(failed),
                None,
            )
            return failed

        if record.state is None:
            return resume_failed("No stored state")

        state = record.state
        if state.runner_type == "flow":
            flow = self._flow_runs.get(report.run_id)
            if flow is None:
                return resume_failed("Flow not registered")
            return flow.resume(report, decision_or_input, stream=stream)
        stream_enabled = stream if stream is not None else state.payload.get("stream", self.stream)
        stream_options = {"include_usage": True} if stream_enabled else None

        def emit(event_type: str, source: str, payload: dict[str, Any]) -> None:
            self._emit(events, report.run_id, event_type, source, payload)

        self._emit(events, report.run_id, "run.resumed", "desk", {})

        worker: Worker | None = None
        if state.runner_type == "worker":
            worker = self._workers.get(state.runner_name)
            if worker is None:
                return resume_failed("Worker not registered")
            updated_report, updated_state = worker.resume(
                adapter=self.adapter,
                state=state,
                decision_or_input=decision_or_input,
                events=events,
                emit=emit,
                temperature=self.temperature,
                max_tokens=self.max_tokens,
                stream=stream_enabled,
                stream_options=stream_options,
                structured_output_retries=self.structured_output_retries,
                max_iterations=self.max_iterations,
                max_tool_calls=self.max_tool_calls,
                model_name=worker.model or self.model,
                respect_context_window=self.respect_context_window,
            )
        elif state.runner_type == "workforce":
            workforce = self._workforces.get(state.runner_name)
            if workforce is None:
                return resume_failed("Workforce not registered")
            updated_report, updated_state = workforce.resume(
                adapter=self.adapter,
                state=state,
                decision_or_input=decision_or_input,
                events=events,
                emit=emit,
                temperature=self.temperature,
                max_tokens=self.max_tokens,
                stream=stream_enabled,
                stream_options=stream_options,
                structured_output_retries=self.structured_output_retries,
                max_iterations=self.max_iterations,
                max_tool_calls=self.max_tool_calls,
                default_model=self.model,
                respect_context_window=self.respect_context_window,
            )
        else:
            return resume_failed("Unknown runner type")

        if updated_state is not None and updated_report.status == "paused":
            updated_state.payload["stream"] = stream_enabled
            self._emit(events, report.run_id, "run.paused", "desk", {})
            self.run_store.update_run(
                report.run_id,
                "paused",
                updated_report.content,
                None,
                updated_state,
            )
        elif updated_report.status == "completed":
            self._emit(events, report.run_id, "run.completed", "desk", {})
            self.run_store.update_run(
                report.run_id,
                "completed",
                updated_report.content,
                self._output_json(updated_report),
                None,
            )
            if worker is not None:
                self._write_memory(worker, updated_report)
        else:
            self._emit(
                events,
                report.run_id,
                "run.failed",
                "desk",
                {"errors": updated_report.errors},
            )
            self.run_store.update_run(
                report.run_id,
                "failed",
                updated_report.content,
                self._output_json(updated_report),
                None,
            )

        return updated_report

    async def aresume(
        self,
        report: Report,
        decision_or_input: Any,
        *,
        stream: bool | None = None,
    ) -> Report:
        record = self.run_store.get_run(report.run_id)
        if record is None:
            failed = Report(
                run_id=report.run_id,
                status="failed",
                content=None,
                reasoning_content=None,
                data=None,
                messages=report.messages,
                tool_calls=report.tool_calls,
                metrics=report.metrics,
                events=report.events,
                pending_action=None,
                errors=["No stored state"],
            )
            return failed
        events = self.run_store.get_events(report.run_id)

        def resume_failed(message: str) -> Report:
            failed = Report(
                run_id=report.run_id,
                status="failed",
                content=None,
                data=None,
                messages=report.messages,
                tool_calls=report.tool_calls,
                metrics=report.metrics,
                events=events,
                pending_action=None,
                errors=[message],
            )
            self._emit(
                events,
                report.run_id,
                "run.failed",
                "desk",
                {"errors": failed.errors},
            )
            self.run_store.update_run(
                report.run_id,
                "failed",
                failed.content,
                self._output_json(failed),
                None,
            )
            return failed

        if record.state is None:
            return resume_failed("No stored state")

        state = record.state
        if state.runner_type == "flow":
            flow = self._flow_runs.get(report.run_id)
            if flow is None:
                return resume_failed("Flow not registered")
            return await flow.aresume(report, decision_or_input, stream=stream)
        stream_enabled = stream if stream is not None else state.payload.get("stream", self.stream)
        stream_options = {"include_usage": True} if stream_enabled else None

        def emit(event_type: str, source: str, payload: dict[str, Any]) -> None:
            self._emit(events, report.run_id, event_type, source, payload)

        self._emit(events, report.run_id, "run.resumed", "desk", {})

        worker: Worker | None = None
        if state.runner_type == "worker":
            worker = self._workers.get(state.runner_name)
            if worker is None:
                return resume_failed("Worker not registered")
            updated_report, updated_state = await worker.aresume(
                adapter=self.adapter,
                state=state,
                decision_or_input=decision_or_input,
                events=events,
                emit=emit,
                temperature=self.temperature,
                max_tokens=self.max_tokens,
                stream=stream_enabled,
                stream_options=stream_options,
                structured_output_retries=self.structured_output_retries,
                max_iterations=self.max_iterations,
                max_tool_calls=self.max_tool_calls,
                model_name=worker.model or self.model,
                respect_context_window=self.respect_context_window,
            )
        elif state.runner_type == "workforce":
            workforce = self._workforces.get(state.runner_name)
            if workforce is None:
                return resume_failed("Workforce not registered")
            updated_report, updated_state = await workforce.aresume(
                adapter=self.adapter,
                state=state,
                decision_or_input=decision_or_input,
                events=events,
                emit=emit,
                temperature=self.temperature,
                max_tokens=self.max_tokens,
                stream=stream_enabled,
                stream_options=stream_options,
                structured_output_retries=self.structured_output_retries,
                max_iterations=self.max_iterations,
                max_tool_calls=self.max_tool_calls,
                default_model=self.model,
                respect_context_window=self.respect_context_window,
            )
        else:
            return resume_failed("Unknown runner type")

        if updated_state is not None and updated_report.status == "paused":
            updated_state.payload["stream"] = stream_enabled
            self._emit(events, report.run_id, "run.paused", "desk", {})
            self.run_store.update_run(
                report.run_id,
                "paused",
                updated_report.content,
                None,
                updated_state,
            )
        elif updated_report.status == "completed":
            self._emit(events, report.run_id, "run.completed", "desk", {})
            self.run_store.update_run(
                report.run_id,
                "completed",
                updated_report.content,
                self._output_json(updated_report),
                None,
            )
            if worker is not None:
                self._write_memory(worker, updated_report)
        else:
            self._emit(
                events,
                report.run_id,
                "run.failed",
                "desk",
                {"errors": updated_report.errors},
            )
            self.run_store.update_run(
                report.run_id,
                "failed",
                updated_report.content,
                self._output_json(updated_report),
                None,
            )

        return updated_report



from typing import Any

MemoryScope = str


class MemoryStore:
    def write(self, key: str, value: Any, scope: MemoryScope) -> None:
        raise NotImplementedError

    def read(self, key: str, scope: MemoryScope) -> Any | None:
        raise NotImplementedError

    def search(self, query: str, scope: MemoryScope) -> list[tuple[str, Any]]:
        raise NotImplementedError

    def reset(self, scope: MemoryScope) -> None:
        raise NotImplementedError
